#include <stdio.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>

/* nom et taille du segment mémoire partagé
   dans notre cas sous linux il apparaitra comme un fichier
   nommé /dev/shmem/shmem_test */
#define SHMEM_NAME "shmem_test"
#define SHMEM_SIZE ( (size_t) 1024)

/* fonctions pour la gestion des erreurs
   elle est basique : une erreur = fin du programme
   clean_exit est un foncion appelée en cas d'erreur
   à la sortie du programme. Elle détruit le segment mémoire
   s'il n'existe pas l'erreur n'est pas grve car on aborte 
   de toute façon */
void handle_error(const char *msg);
void clean_exit(void);

/* Pour créer un map mémoire */
void map_new(size_t size, void **address, int *pfd);

/* Fonctions de test d'écriture et de lecture */
void write_mem(char *buffer, size_t size);
void print_mem(char *buffer, size_t size);

void write_file(int fd, size_t size);
size_t print_file(int fd, size_t size);

/* **************************** */

void handle_error(const char *msg)
{
  perror(msg);
  exit(EXIT_FAILURE);
}

void clean_exit(void)
{
  shm_unlink(SHMEM_NAME);
}

void map_new(size_t size, void **address, int *pfd)
{
  assert( address != NULL );
  assert( pfd != NULL );
  assert( size > 0 );

  /* On crée un nouvel objet mémoire */
  if ( (*pfd = shm_open(SHMEM_NAME, O_RDWR | O_CREAT, 0600)) == -1 )
    handle_error("shm_open");

  /* On lui assigne une taille */
  if ( lseek(*pfd, size-1, SEEK_SET) == -1 )
    handle_error("lseek");

  if ( write(*pfd, "", 1) == -1 )
    handle_error("lseek/write");

  /* On mmape */
  if ( (*address = mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_SHARED, *pfd, 0)) == (void*) -1 )
    handle_error("mmap");

}

void write_mem(char *buffer, size_t size)
{
  size_t i;
  for(i=0; i<size; ++i)
    buffer[i]='A'+(unsigned char) (i % 26);
  msync(buffer, size, MS_SYNC);
}

void print_mem(char *buffer, size_t size)
{
  printf("%.*s", (int) size, buffer);
}

void write_file(int fd, size_t size)
{
  size_t i;
  char c;
  lseek(fd, SEEK_SET, 0);
  for(i=0; i<size; ++i) {
    c='0' + (unsigned char)(i %10);
    if (write(fd, &c, 1) == -1)
      handle_error("write");
  }
}

size_t print_file(int fd, size_t size)
{
  char buffer[200];
  ssize_t byte_read;
  off_t offset;

  offset=0;
  while ( (byte_read=pread(fd, buffer, 200, offset)) > 0 ) {
    offset+=byte_read;
    if (offset>size)
      byte_read-=offset-size;
    printf("%.*s", (int) byte_read, buffer);
  }
  if (byte_read==-1)
    handle_error("pread");
  return (size_t) offset;
}

int main(void)
{

  char *mem;
  int fd;
  size_t sz;

  /* On s'assure que le sement mémoire partagé est libéré
     quelle que soit la situation de sorite. */
  atexit(clean_exit);

  /* On crée le map mémoire et le fichier */
  map_new(SHMEM_SIZE, (void**) &mem, &fd);

  printf("On écrit en mémoire l'alphabet en boucle sur %zu octets\n\n", SHMEM_SIZE);
  write_mem(mem, SHMEM_SIZE);

  printf("On lit le fd :\n");
  sz=print_file(fd,SHMEM_SIZE);
  printf("\n\t%zu octets lus\n\n", sz);

  printf("On écrit dans le fichier les chiffres de 0 à 9 en boucle sur %zu octets\n\n", SHMEM_SIZE);
  write_file(fd, SHMEM_SIZE);

  printf("On lit la mémoire sur %zu octets :\n", SHMEM_SIZE);
  print_mem(mem, SHMEM_SIZE);
  printf("\n");

  /* On sort proprement */
  munmap(mem, SHMEM_SIZE);
  close(fd);
  shm_unlink(SHMEM_NAME);
  
  return 0;
}
